package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.Block;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.block.entity.worker.WorkerState;
import ru.dbotthepony.mc.otm.capability.*;
import ru.dbotthepony.mc.otm.capability.matter.IMatterGridCell;
import ru.dbotthepony.mc.otm.capability.matter.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.matter.MatterHandlerCapability;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.container.MatteryContainerHandler;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.menu.MatterBottlerMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityMatterBottler extends BlockEntityMatteryPowered implements IMatterGridCell {
	public BlockEntityMatterBottler(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_BOTTLER, p_155229_, p_155230_);
		energy = new MatteryMachineEnergyStorage(this, MatteryMachineEnergyStorage.MachineType.WORKER);
	}

	private static final TranslatableComponent MACHINE_NAME = new TranslatableComponent("block.overdrive_that_matters.matter_bottler");

	public boolean getWorkFlow() {
		return work_flow;
	}

	public void setWorkFlow(boolean work_flow) {
		this.work_flow = work_flow;
		this.setChanged();
	}

	public void switchWorkFlow() {
		setWorkFlow(!work_flow);
	}

	// true - bottling
	// false - unbottling
	protected boolean work_flow = true;

	// true - continue even if empty when bottling / if full while unbottling
	// false - spit into output slot
	protected boolean work_behavior = true;

	public final MatteryContainer work_slots = new MatteryContainer(this::setChanged, 6) {
		@Override
		public int getMaxStackSize(int slot) {
			return 1;
		}
	};

	public final MatteryContainerHandler work_slots_handler = work_slots.handler((slot, stack) -> {
		if (work_flow) {
			return slot < 3 && stack.getCapability(MatteryCapability.MATTER).isPresent();
		}

		return slot >= 3 && stack.getCapability(MatteryCapability.MATTER).isPresent();
	}, (slot, amount, stack) -> {
		if (work_flow) {
			return slot >= 3;
		}

		return slot < 3;
	});

	public final MatterHandlerCapability matter = new MatterHandlerCapability(this::setChanged, IMatterHandler.MatterDirection.BIDIRECTIONAL, new Fraction(4)) {
		@Nonnull
		@Override
		public MatterDirection getDirection() {
			return work_flow ? MatterDirection.RECEIVE : MatterDirection.EXTRACT;
		}
	};

	private static final Fraction MATTER_EXCHANGE_RATE = new Fraction("0.04");
	private static final Fraction ENERGY_CONSUMPTION = new Fraction(20);

	private LazyOptional<IMatterGridCell> cell_resolver = LazyOptional.of(() -> this);

	protected Fraction initial_capacity;
	protected ItemStack last_work_stack;

	private MatterGrid grid;

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return matter;
	}

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			scheduleDiscoverNeighbours(p_155231_, getBlockPos());
	}

	private boolean valid = true;

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
		work_slots_handler.invalidate();
		matter.invalidate();
		cell_resolver.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
		work_slots_handler.revive();
		matter.revive();
		cell_resolver = LazyOptional.of(() -> this);
	}

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid) {
			if (cap == MatteryCapability.MATTER) {
				return matter.get().cast();
			}

			if (cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) {
				return work_slots_handler.get().cast();
			}

			if (cap == MatteryCapability.MATTER_CELL) {
				return cell_resolver.cast();
			}
		}

		return super.getCapability(cap, side);
	}

	@Override
	protected Component getDefaultDisplayName() {
		return MACHINE_NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterBottlerMenu(containerID, inventory, this);
	}

	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		nbt.put("work_slots", work_slots.serializeNBT());
		nbt.putBoolean("work_flow", work_flow);
		nbt.put("matter_capability", matter.serializeNBT());
	}

	@Override
	public void load(CompoundTag nbt) {
		super.load(nbt);
		work_slots.deserializeNBT(nbt.get("work_slots"));
		work_flow = nbt.getBoolean("work_flow");

		if (nbt.get("matter_capability") instanceof CompoundTag _matter)
			matter.deserializeNBT(_matter);
	}

	private static final Fraction EXTRACTION_TICKS = new Fraction(200);

	public float getWorkProgress() {
		if (last_work_stack == null) {
			return 0;
		}

		var cap = last_work_stack.getCapability(MatteryCapability.MATTER).resolve();

		if (cap.isEmpty()) {
			return 0;
		}

		if (work_flow) {
			if (cap.get().getMaxStoredMatter().minus(initial_capacity).compareTo(Fraction.ZERO) <= 0) {
				return 0;
			}

			return cap.get().getStoredMatter().minus(initial_capacity).div(cap.get().getMaxStoredMatter().minus(initial_capacity)).toFloat();
		}

		if (initial_capacity.compareTo(Fraction.ZERO) <= 0) {
			return 0;
		}

		return Fraction.ONE.minus(cap.get().getStoredMatter().div(initial_capacity)).toFloat();
	}

	public static <T extends BlockEntity> void tick(Level level, BlockPos blockPos, BlockState blockState, T t) {
		if (t instanceof BlockEntityMatterBottler tile) {
			tile.batteryChargeLoop();

			if (tile.isBlockedByRedstone()) {
				if (tile.getBlockState().getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.IDLE) {
					level.setBlock(tile.getBlockPos(), tile.getBlockState().setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
				}

				return;
			}

			ItemStack work_stack = null;
			IMatterHandler capability = null;
			int align = tile.work_flow ? 0 : 3;
			int work_slot = -1;

			for (int i = align; i < align + 3; i++) {
				var get = tile.work_slots.getItem(i);

				if (!get.isEmpty()) {
					var cap = get.getCapability(MatteryCapability.MATTER).resolve();

					if (cap.isPresent() && cap.get().getDirection() != (tile.work_flow ? IMatterHandler.MatterDirection.EXTRACT : IMatterHandler.MatterDirection.RECEIVE) && (tile.work_flow ? cap.get().getMissingMatter().compareTo(Fraction.ZERO) > 0 : cap.get().getStoredMatter().compareTo(Fraction.ZERO) > 0)) {
						work_stack = get;
						capability = cap.get();
						work_slot = i;
						break;
					}
				}
			}

			if (work_stack == null) {
				tile.last_work_stack = null;
				tile.initial_capacity = null;
			} else if (work_stack != tile.last_work_stack) {
				tile.last_work_stack = work_stack;
				tile.initial_capacity = capability.getStoredMatter();
			}

			if (tile.work_flow) {
				if (capability != null) {
					if (tile.getBlockState().getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.WORKING) {
						level.setBlock(tile.getBlockPos(), tile.getBlockState().setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS);
					}

					if (tile.matter.getStoredMatter().compareTo(MATTER_EXCHANGE_RATE) < 0) {
						if (tile.grid != null) {
							var extracted = tile.grid.extractMatter(
								tile.matter.getMissingMatter()
									.min(MATTER_EXCHANGE_RATE.times(EXTRACTION_TICKS))
									.min(capability.getMissingMatter().minus(tile.matter.getStoredMatter()))
								, true);

							if (extracted.compareTo(Fraction.ZERO) > 0) {
								var received = tile.matter.receiveMatterOuter(extracted, false);
								tile.grid.extractMatter(received, false);
							}
						}
					}

					if (tile.matter.getStoredMatter().compareTo(Fraction.ZERO) > 0) {
						var energy = tile.energy.extractEnergyInner(ENERGY_CONSUMPTION, true);

						if (energy.compareTo(Fraction.ZERO) > 0) {
							var matter = capability.receiveMatterOuter(MATTER_EXCHANGE_RATE.min(tile.matter.getStoredMatter()).times(energy.div(ENERGY_CONSUMPTION)), true);

							if (matter.compareTo(Fraction.ZERO) > 0) {
								tile.energy.extractEnergyInner(ENERGY_CONSUMPTION.times(matter.div(MATTER_EXCHANGE_RATE)), false);
								capability.receiveMatterOuter(matter, false);
								tile.matter.extractMatterInner(matter, false);

								if (capability.getMissingMatter().compareTo(Fraction.ZERO) == 0) {
									for (int i = 3; i < 6; i++) {
										if (tile.work_slots.getItem(i).isEmpty()) {
											tile.work_slots.setItem(work_slot, ItemStack.EMPTY);
											tile.work_slots.setItem(i, work_stack);
											break;
										}
									}
								}
							}
						}
					}
				} else {
					if (tile.getBlockState().getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.IDLE) {
						level.setBlock(tile.getBlockPos(), tile.getBlockState().setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
					}
				}
			} else {
				if (capability != null) {
					if (tile.getBlockState().getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.WORKING) {
						level.setBlock(tile.getBlockPos(), tile.getBlockState().setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS);
					}

					var energy = tile.energy.extractEnergyInner(ENERGY_CONSUMPTION, true);

					if (energy.compareTo(Fraction.ZERO) > 0) {
						var matter = capability.extractMatterOuter(MATTER_EXCHANGE_RATE.min(tile.matter.getMissingMatter()).times(energy.div(ENERGY_CONSUMPTION)), true);

						if (matter.compareTo(Fraction.ZERO) > 0) {
							tile.energy.extractEnergyInner(ENERGY_CONSUMPTION.times(matter.div(MATTER_EXCHANGE_RATE)), false);
							capability.extractMatterOuter(matter, false);
							tile.matter.receiveMatterInner(matter, false);

							if (capability.getStoredMatter().compareTo(Fraction.ZERO) == 0) {
								for (int i = 2; i >= 0; i--) {
									if (tile.work_slots.getItem(i).isEmpty()) {
										tile.work_slots.setItem(work_slot, ItemStack.EMPTY);
										tile.work_slots.setItem(i, work_stack);
										break;
									}
								}
							}
						}
					}
				} else {
					if (tile.getBlockState().getValue(WorkerState.SEMI_WORKER_STATE) != WorkerState.IDLE) {
						level.setBlock(tile.getBlockPos(), tile.getBlockState().setValue(WorkerState.SEMI_WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS);
					}
				}

				if (tile.matter.getStoredMatter().compareTo(Fraction.ZERO) > 0 && tile.grid != null) {
					var diff = tile.matter.extractMatterInner(tile.matter.getStoredMatter(), true);
					var diff2 = tile.grid.softPushMatter(diff, true);

					tile.matter.extractMatterInner(diff2, false);
					tile.grid.softPushMatter(diff2, false);
				}
			}
		}
	}
}
