package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.entity.BlockEntity;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import ru.dbotthepony.mc.otm.block.entity.worker.BlockEntityMatteryWorker;
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJob;
import ru.dbotthepony.mc.otm.block.entity.worker.MachineJobStatus;
import ru.dbotthepony.mc.otm.capability.*;
import ru.dbotthepony.mc.otm.capability.matter.*;
import ru.dbotthepony.mc.otm.container.MatteryContainerHandler;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.MatterScannerMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.Collection;
import java.util.UUID;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class BlockEntityMatterScanner extends BlockEntityMatteryWorker implements IMatterGridCell {
	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.matter_scanner");

	public final MatteryContainer input_slot = new MatteryContainer(this::setChanged, 1);

	private final MatteryContainerHandler input_handler = input_slot.handler(
		(slot, stack) -> MatterRegistry.canDecompose(stack),
		(slot, amount, stack) -> is_idling
	);

	public BlockEntityMatterScanner(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_SCANNER, p_155229_, p_155230_);
		energy = new MatteryMachineEnergyStorage(this, MatteryMachineEnergyStorage.MachineType.WORKER, new Fraction(40000), new Fraction(400), new Fraction(400));
	}

	private boolean valid = true;

	private LazyOptional<IMatterGridCell> resolver_grid = LazyOptional.of(() -> this);

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid) {
			if (cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
				return input_handler.get().cast();

			if (cap == MatteryCapability.MATTER_CELL)
				return resolver_grid.cast();
		}

		return super.getCapability(cap, side);
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
		input_handler.invalidate();
		resolver_grid.invalidate();
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
		input_handler.revive();
		resolver_grid = LazyOptional.of(() -> this);
	}

	@Override
	public void onPatternAdded(PatternState state) {
		is_idling = false;
	}

	@Override
	public void onPatternRemoved(PatternState state) {
		is_idling = false;
	}

	@Override
	public void onPatternUpdated(PatternState new_state, PatternState old_state) {
		is_idling = false;
	}

	@Override
	public void setRemoved() {
		super.setRemoved();

		if (level != null && !level.isClientSide && grid != null)
			grid.remove(this);
	}

	@Override
	public void onNeighbourMatterCell(Level level, BlockPos pos, Direction direction, IMatterGridCell cell, BlockEntity entity) {
		is_idling = false;
	}

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterScannerMenu(containerID, inventory, this);
	}

	@Override
	public void saveAdditional(CompoundTag nbt) {
		super.saveAdditional(nbt);
		nbt.put("work_slots", input_slot.serializeNBT());
	}

	@Override
	public void load(CompoundTag nbt) {
		input_slot.deserializeNBT(nbt.get("work_slots"));
		super.load(nbt);
	}

	private static final Fraction BASE_CONSUMPTION = new Fraction("40");

	@Override
	@Nonnull
	protected Fraction getBaseConsumption() {
		return BASE_CONSUMPTION;
	}

	@Nonnull
	@Override
	protected MachineJobStatus onJobFinish(MachineJob job) {
		if (grid == null)
			return new MachineJobStatus(false, 100);

		ItemStack stack = job.stack();

		if (stack.isEmpty() || !MatterRegistry.hasMatterValue(stack))
			return new MachineJobStatus();


		Collection<PatternState> get_state = grid.findPatterns(stack.getItem());
		PatternState find_state = null;

		for (PatternState state : get_state) {
			if (state.item() == stack.getItem()) {
				if (find_state == null && state.research_percent() < 1d) {
					find_state = state;
				} else if (find_state != null && find_state.research_percent() < state.research_percent()) {
					find_state = state;
				}
			}
		}

		PatternState new_state;

		if (find_state != null) {
			new_state = new PatternState(find_state.id(), stack.getItem(), find_state.research_percent() + 0.2d);
		} else {
			new_state = new PatternState(UUID.randomUUID(), stack.getItem(), 0.2d);
		}

		if (grid.insertPattern(new_state, false, false).status() != PatternInsertStatus.Status.FAIL) {
			return new MachineJobStatus();
		}

		return new MachineJobStatus(false, 200);
	}

	@Nullable
	@Override
	protected MachineJob getNextJob() {
		if (grid == null)
			return null;

		ItemStack stack = input_slot.getItem(0);

		if (stack.isEmpty() || !MatterRegistry.canDecompose(stack))
			return null;

		Collection<PatternState> get_state = grid.findPatterns(stack.getItem());
		PatternState find_state = null;

		for (PatternState state : get_state) {
			if (state.item() == stack.getItem() && state.research_percent() < 1d) {
				find_state = state;
			} else if (state.item() == stack.getItem() && state.research_percent() >= 1d) {
				return null;
			}
		}

		PatternState new_state;

		if (find_state != null) {
			new_state = new PatternState(find_state.id(), stack.getItem(), find_state.research_percent() + 0.2d);
		} else {
			new_state = new PatternState(UUID.randomUUID(), stack.getItem(), 0.2d);
		}

		if (grid.insertPattern(new_state, false, true).status() != PatternInsertStatus.Status.FAIL) {
			ItemStack copy = stack.copy();
			copy.setCount(1);
			stack.shrink(1);
			input_slot.setChanged();
			return new MachineJob(copy, MatterRegistry.getMatterValue(copy).toDouble() * 35_000d);
		}

		return null;
	}

	@Nullable
	@Override
	public IPatternStorage getPatternStorage() {
		return null;
	}

	private MatterGrid grid;

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			scheduleDiscoverNeighbours(p_155231_, getBlockPos());
	}

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return null;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}
}
