package ru.dbotthepony.mc.otm.capability;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraftforge.common.util.LazyOptional;
import ru.dbotthepony.mc.otm.OverdriveThatMatters;
import ru.dbotthepony.mc.otm.storage.IStorageIdentity;
import ru.dbotthepony.mc.otm.storage.IStorageStack;
import ru.dbotthepony.mc.otm.storage.StorageGrid;

import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.function.Function;
import java.util.function.Supplier;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class AbstractStorageGridCell implements IStorageGridCell {
	private LazyOptional<IStorageGridCell> resolver = LazyOptional.of(() -> this);
	private boolean valid = true;
	protected final ArrayList<IStorageIdentity<?>> components = new ArrayList<>();

	@Override
	public List<? extends IStorageIdentity<?>> getComponents() {
		return Collections.unmodifiableList(components);
	}

	@SuppressWarnings("unchecked")
	public <T extends IStorageStack, U extends IStorageIdentity<T>> U computeIfAbsent(Class<T> identity, Supplier<U> provider) {
		for (var component : components) {
			if (component.storageIdentity() == identity) {
				return (U) component;
			}
		}

		var factory = provider.get();
		addStorageComponent(factory);
		return factory;
	}

	@SuppressWarnings("unchecked")
	public <T extends IStorageStack, U extends IStorageIdentity<T>> U computeIfAbsent(Class<T> identity, Function<Class<T>, U> provider) {
		for (var component : components) {
			if (component.storageIdentity() == identity) {
				return (U) component;
			}
		}

		var factory = provider.apply(identity);
		addStorageComponent(factory);
		return factory;
	}

	public void addStorageComponent(IStorageIdentity<?> component) {
		for (var component1 : components) {
			if (component == component1 || component1.storageIdentity() == component.storageIdentity()) {
				return;
			}
		}

		detachStorage();
		components.add(component);
		attachStorage();
	}

	public void removeStorageComponent(IStorageIdentity<?> component) {
		for (var component1 : components) {
			if (component == component1 || component1.storageIdentity() == component.storageIdentity()) {
				detachStorage();
				components.remove(component1);
				attachStorage();
				return;
			}
		}
	}

	public void invalidate() {
		if (!valid)
			return;

		valid = false;
		resolver.invalidate();
		// detachStorage();
	}

	public void revive() {
		if (valid)
			return;

		valid = true;
		resolver = LazyOptional.of(() -> this);
		// attachStorage();
	}

	public LazyOptional<IStorageGridCell> get() {
		return valid ? resolver : LazyOptional.empty();
	}

	protected StorageGrid storage_grid;

	@Nullable
	@Override
	public StorageGrid getStorageGrid() {
		return storage_grid;
	}

	@Override
	public void setStorageGrid(@Nullable StorageGrid grid) {
		storage_grid = grid;
	}

	@Override
	public boolean isValidStorageCell() {
		return valid;
	}
}
