package ru.dbotthepony.mc.otm.capability.android;

import net.minecraft.MethodsReturnNonnullByDefault;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.server.level.ServerPlayer;
import net.minecraft.world.effect.MobEffect;
import net.minecraft.world.effect.MobEffects;
import net.minecraft.world.entity.LivingEntity;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.INBTSerializable;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.energy.IEnergyStorage;
import net.minecraftforge.event.entity.living.LivingEvent;
import net.minecraftforge.event.entity.living.LivingHurtEvent;
import net.minecraftforge.eventbus.api.EventPriority;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.fml.event.lifecycle.FMLCommonSetupEvent;
import net.minecraftforge.network.PacketDistributor;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.android.AndroidFeature;
import ru.dbotthepony.mc.otm.android.AndroidFeatureType;
import ru.dbotthepony.mc.otm.capability.IMatteryEnergyStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.core.Fraction;
import ru.dbotthepony.mc.otm.network.MatteryNetworking;
import ru.dbotthepony.mc.otm.network.android.AndroidBatteryPacket;
import ru.dbotthepony.mc.otm.network.android.AndroidEnergyPacket;
import ru.dbotthepony.mc.otm.network.android.AndroidFeaturePacket;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;
import java.math.BigDecimal;
import java.util.*;

@MethodsReturnNonnullByDefault
@ParametersAreNonnullByDefault
public class AndroidCapability implements ICapabilityProvider, IAndroidCapability, INBTSerializable<CompoundTag> {
	protected final LivingEntity ent;
	protected Fraction energy_stored = Fraction.ZERO;
	protected Fraction energy_stored_max = new Fraction(60_000);
	protected ItemStack battery = ItemStack.EMPTY;

	private Fraction network_energy = new Fraction(-1);
	private Fraction network_energy_max = new Fraction(-1);
	private ItemStack network_battery = ItemStack.EMPTY;

	protected final Map<AndroidFeatureType<?>, AndroidFeature> features = new HashMap<>();
	protected final ArrayList<Object> network_queue = new ArrayList<>();

	protected final ArrayList<Runnable> delayed_tick_server = new ArrayList<>();

	protected boolean network_first = false;

	protected boolean addFeature(AndroidFeature feature) {
		if (features.containsKey(feature.type))
			return false;

		features.put(feature.type, feature);

		if (!ent.level.isClientSide) {
			delayed_tick_server.add(feature::applyModifiers);
		}

		if (ent instanceof ServerPlayer ply) {
			sendNetwork(new AndroidFeaturePacket(true, feature));
		}

		return true;
	}

	@Override
	@SuppressWarnings("unchecked")
	public <T extends AndroidFeature> T addFeature(@Nonnull AndroidFeatureType<T> feature) {
		var get = features.get(feature);

		if (get != null)
			return (T) get;

		var factory = feature.factory(this);
		features.put(feature, factory);

		if (!ent.level.isClientSide) {
			delayed_tick_server.add(factory::applyModifiers);
		}

		if (ent instanceof ServerPlayer ply) {
			sendNetwork(new AndroidFeaturePacket(true, factory));
		}

		return factory;
	}

	@Override
	public boolean removeFeature(@Nonnull AndroidFeatureType<?> feature) {
		var removed = features.remove(feature);

		if (removed != null) {
			if (!ent.level.isClientSide) {
				delayed_tick_server.add(removed::removeModifiers);
			}

			if (ent instanceof ServerPlayer ply) {
				sendNetwork(new AndroidFeaturePacket(false, removed));
			}

			return true;
		}

		return false;
	}

	@Override
	public boolean hasFeature(@Nullable AndroidFeatureType<?> feature) {
		return features.containsKey(feature);
	}

	@Override
	public boolean hasFeatureLevel(@Nullable AndroidFeatureType<?> feature, int level) {
		var get = features.get(feature);

		if (get == null) {
			return false;
		}

		return get.getLevel() >= level;
	}

	@Nullable
	@SuppressWarnings("unchecked")
	public <T extends AndroidFeature> T getFeature(AndroidFeatureType<T> feature) {
		return (T) features.get(feature);
	}

	public static final Set<MobEffect> UNAFFECTED_EFFECTS = new HashSet<>();

	public static void registerEffects(final FMLCommonSetupEvent event) {
		UNAFFECTED_EFFECTS.add(MobEffects.CONDUIT_POWER);
		UNAFFECTED_EFFECTS.add(MobEffects.HEAL);
		// maybe it makes things go haywire idk
		// UNAFFECTED_EFFECTS.add(MobEffects.HARM);
		UNAFFECTED_EFFECTS.add(MobEffects.REGENERATION);
		UNAFFECTED_EFFECTS.add(MobEffects.WATER_BREATHING);
		UNAFFECTED_EFFECTS.add(MobEffects.POISON);
		// even skeletons can be withered
		// UNAFFECTED_EFFECTS.add(MobEffects.WITHER);
		UNAFFECTED_EFFECTS.add(MobEffects.HEALTH_BOOST);
		UNAFFECTED_EFFECTS.add(MobEffects.ABSORPTION);
		UNAFFECTED_EFFECTS.add(MobEffects.SATURATION);
		UNAFFECTED_EFFECTS.add(MobEffects.DOLPHINS_GRACE);
	}

	@Override
	public void invalidateNetworkState() {
		network_energy = Fraction.MINUS_ONE;
		network_energy_max = Fraction.MINUS_ONE;
		network_battery = ItemStack.EMPTY;

		if (ent instanceof ServerPlayer ply) {
			for (var feature : features.values()) {
				sendNetwork(new AndroidFeaturePacket(true, feature));
				feature.invalidateNetwork();
			}
		}
	}

	@Override
	public void setEnergy(Fraction value) {
		energy_stored = value;
	}

	@Override
	public void setMaxEnergy(Fraction value) {
		energy_stored_max = value;
	}

	public AndroidCapability(LivingEntity ent) {
		this.ent = ent;
	}

	@SubscribeEvent
	@SuppressWarnings("unused")
	public static void onLivingTick(LivingEvent.LivingUpdateEvent event) {
		var ent = event.getEntity();
		ent.getCapability(MatteryCapability.ANDROID).ifPresent(ent.level.isClientSide ? IAndroidCapability::tickClient : IAndroidCapability::tick);
	}

	@SubscribeEvent(priority = EventPriority.LOWEST)
	@SuppressWarnings("unused")
	public static void onHurtEvent(LivingHurtEvent event) {
		if (event.isCanceled())
			return;

		var ent = event.getEntity();
		ent.getCapability(MatteryCapability.ANDROID).ifPresent((cap) -> cap.onHurt(event));
	}

	public void onHurt(LivingHurtEvent event) {
		for (var feature : features.values()) {
			feature.onHurt(event);
		}
	}

	@Override
	@Nonnull
	public CompoundTag serializeNBT() {
		CompoundTag tag = new CompoundTag();
		tag.putString("energy_stored", energy_stored.toString());
		tag.putString("energy_stored_max", energy_stored_max.toString());
		tag.put("battery", battery.serializeNBT());

		var features_nbt = new ListTag();

		for (var feature : features.values()) {
			var compound = new CompoundTag();
			feature.serializeNBT(compound);
			compound.putString("id", Objects.requireNonNull(feature.type.getRegistryName()).toString());
			features_nbt.add(compound);
		}

		tag.put("features", features_nbt);

		return tag;
	}

	@Override
	public void deserializeNBT(CompoundTag compound) {
		if (compound.contains("energy_stored"))
			energy_stored = Fraction.deserializeNBT(compound.get("energy_stored"));

		if (compound.contains("energy_stored_max"))
			energy_stored_max = Fraction.deserializeNBT(compound.get("energy_stored_max"));

		if (compound.contains("battery"))
			battery = ItemStack.of(compound.getCompound("battery"));

		features.clear();

		var features_nbt = compound.getList("features", Tag.TAG_COMPOUND);

		for (var _tag : features_nbt) {
			if (_tag instanceof CompoundTag tag) {
				var get_feature = Registry.ANDROID_FEATURES.getValue(new ResourceLocation(tag.getString("id")));

				if (get_feature != null && get_feature.isApplicable(this)) {
					var feature = get_feature.factory(this);
					feature.deserializeNBT(tag);

					addFeature(feature);

					if (ent.level == null || !ent.level.isClientSide) {
						delayed_tick_server.add(feature::applyModifiers);
					}
				}
			}
		}
	}

	public void dropBattery() {
		if (battery.isEmpty())
			return;

		ent.spawnAtLocation(battery);
		battery = ItemStack.EMPTY;
	}

	protected void sendNetwork(Object packet) {
		if (ent instanceof ServerPlayer ply) {
			if (network_first) {
				MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with(() -> ply), packet);
			} else {
				network_queue.add(packet);
			}
		}
	}

	protected void tickNetwork() {
		network_first = true;

		if (ent instanceof ServerPlayer ply) {
			if (!energy_stored.equals(network_energy)) {
				network_energy = energy_stored;
				sendNetwork(new AndroidEnergyPacket(false, energy_stored));
			}

			if (!energy_stored_max.equals(network_energy_max)) {
				network_energy_max = energy_stored_max;
				sendNetwork(new AndroidEnergyPacket(true, energy_stored_max));
			}

			if (!network_battery.equals(battery, false)) {
				network_battery = battery.copy();
				sendNetwork(new AndroidBatteryPacket(battery));
			}

			if (network_queue.size() != 0) {
				for (var packet : network_queue) {
					MatteryNetworking.CHANNEL.send(PacketDistributor.PLAYER.with(() -> ply), packet);
				}

				network_queue.clear();
			}
		}
	}

	protected void tickInnerClient() {

	}

	protected void tickInnerClientAlways() {

	}

	@Override
	public void tickClient() {
		delayed_tick_server.clear();

		if (!ent.isAlive())
			return;

		tickInnerClientAlways();

		if (isAndroid()) {
			tickInnerClient();

			for (var feature : features.values()) {
				feature.tickClient();
			}
		}
	}

	@Override
	public void tick() {
		if (!ent.isAlive())
			return;

		tickServerAlways();

		if (isAndroid()) {
			tickServer();

			for (var feature : features.values()) {
				feature.tickServer();
			}
		}

		for (var runnable : delayed_tick_server) {
			runnable.run();
		}

		delayed_tick_server.clear();
		tickNetwork();
	}

	protected void tickServerAlways() {

	}

	protected void tickServer() {
		if (ent.getAirSupply() < ent.getMaxAirSupply())
			ent.setAirSupply(ent.getMaxAirSupply());

		for (var effect : UNAFFECTED_EFFECTS)
			if (ent.hasEffect(effect))
				ent.removeEffect(effect);

		if (!battery.isEmpty()) {
			Fraction demand = energy_stored_max.minus(energy_stored);

			if (demand.compareTo(Fraction.ZERO) > 0) {
				Optional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY).resolve();

				if (get_mattery.isPresent()) {
					demand = demand.minus(get_mattery.get().extractEnergyInner(demand, false));
				}

				if (demand.compareTo(Fraction.ONE) >= 0) {
					Optional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY).resolve();

					if (get_energy.isPresent()) {
						demand = demand.minus(MatteryCapability.drainFE(get_energy.get(), demand, false));
					}
				}

				energy_stored = energy_stored_max.minus(demand);
			}
		}
	}

	@Nonnull
	@Override
	public Fraction extractEnergyOuter(Fraction howMuch, boolean simulate) {
		return Fraction.ZERO;
	}

	@Nonnull
	@Override
	public Fraction extractEnergyInner(Fraction howMuch, boolean simulate) {
		Fraction drained = Fraction.ZERO;

		if (battery != ItemStack.EMPTY) {
			Optional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY).resolve();

			if (get_mattery.isPresent()) {
				Fraction changed = get_mattery.get().extractEnergyOuter(howMuch, simulate);

				if (changed.compareTo(Fraction.ZERO) > 0) {
					drained = drained.plus(changed);
					howMuch = howMuch.minus(changed);

					if (howMuch.compareTo(Fraction.ZERO) <= 0) {
						if (!simulate && ent instanceof ServerPlayer ply) {
							ply.awardStat(Registry.Names.POWER_CONSUMED, drained.toInt() * 10);
						}

						return drained;
					}
				}
			}

			if (howMuch.compareTo(Fraction.ONE) >= 0) {
				Optional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY).resolve();

				if (get_energy.isPresent()) {
					Fraction changed = MatteryCapability.drainFE(get_energy.get(), howMuch, simulate);

					if (changed.compareTo(Fraction.ZERO) > 0) {
						drained = drained.plus(changed);
						howMuch = howMuch.minus(changed);

						if (howMuch.compareTo(Fraction.ZERO) <= 0) {
							if (!simulate && ent instanceof ServerPlayer ply) {
								ply.awardStat(Registry.Names.POWER_CONSUMED, drained.toInt() * 10);
							}

							return drained;
						}
					}
				}
			}
		}

		Fraction new_energy = energy_stored.minus(howMuch).max(Fraction.ZERO);
		drained = drained.plus(energy_stored.minus(new_energy));

		if (!simulate) {
			energy_stored = new_energy;

			if (ent instanceof ServerPlayer ply) {
				ply.awardStat(Registry.Names.POWER_CONSUMED, drained.toInt() * 10);
			}
		}

		return drained;
	}

	@Nonnull
	@Override
	public Fraction receiveEnergyOuter(Fraction howMuch, boolean simulate) {
		Fraction received = Fraction.ZERO;

		if (battery != ItemStack.EMPTY) {
			Optional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY).resolve();

			if (get_mattery.isPresent()) {
				Fraction changed = get_mattery.get().receiveEnergyOuter(howMuch, simulate);

				if (changed.compareTo(Fraction.ZERO) > 0) {
					received = received.plus(changed);
					howMuch = howMuch.minus(changed);

					if (howMuch.compareTo(Fraction.ZERO) <= 0) {
						return received;
					}
				}
			}

			if (howMuch.compareTo(Fraction.ONE) >= 0) {
				Optional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY).resolve();

				if (get_energy.isPresent()) {
					Fraction changed = MatteryCapability.floodFE(get_energy.get(), howMuch, simulate);

					if (changed.compareTo(Fraction.ZERO) > 0) {
						received = received.plus(changed);
						howMuch = howMuch.minus(changed);

						if (howMuch.compareTo(Fraction.ZERO) <= 0) {
							return received;
						}
					}
				}
			}
		}

		Fraction new_energy = energy_stored.plus(howMuch).min(energy_stored_max);
		received = received.plus(new_energy.minus(energy_stored));

		if (!simulate) {
			energy_stored = new_energy;
		}

		return received;
	}

	@Nonnull
	@Override
	public Fraction receiveEnergyInner(Fraction howMuch, boolean simulate) {
		return receiveEnergyOuter(howMuch, simulate);
	}

	@Override
	public LivingEntity getEntity() {
		return this.ent;
	}

	@Nonnull
	@Override
	public ItemStack getBatteryItemStack() {
		return battery;
	}

	@Override
	public void setBatteryItemStack(@Nonnull ItemStack stack) {
		battery = stack;
	}

	@Override
	@Nonnull
	public Fraction getBatteryLevel() {
		if (battery != ItemStack.EMPTY) {
			Optional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY).resolve();

			if (get_mattery.isPresent()) {
				return get_mattery.get().getBatteryLevel().plus(energy_stored);
			}

			Optional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY).resolve();

			if (get_energy.isPresent()) {
				return energy_stored.plus(get_energy.get().getEnergyStored());
			}
		}

		return energy_stored;
	}

	@Override
	@Nonnull
	public Fraction getMaxBatteryLevel() {
		if (battery != ItemStack.EMPTY) {
			Optional<IMatteryEnergyStorage> get_mattery = battery.getCapability(MatteryCapability.ENERGY).resolve();

			if (get_mattery.isPresent()) {
				return get_mattery.get().getMaxBatteryLevel().plus(energy_stored_max);
			}

			Optional<IEnergyStorage> get_energy = battery.getCapability(CapabilityEnergy.ENERGY).resolve();

			if (get_energy.isPresent()) {
				return energy_stored_max.plus(get_energy.get().getMaxEnergyStored());
			}
		}

		return energy_stored_max;
	}

	private final LazyOptional<IAndroidCapability> resolver = LazyOptional.of(() -> this).cast();

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (cap == MatteryCapability.ANDROID) {
			return resolver.cast();
		}

		return LazyOptional.empty();
	}
}
