package ru.dbotthepony.mc.otm.client.screen.panels;

import com.mojang.blaze3d.systems.RenderSystem;
import com.mojang.blaze3d.vertex.PoseStack;
import net.minecraft.client.Minecraft;
import net.minecraft.client.gui.Font;
import net.minecraft.client.renderer.GameRenderer;
import net.minecraft.network.chat.Component;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.client.gui.GuiUtils;
import ru.dbotthepony.mc.otm.client.screen.MatteryScreen;
import ru.dbotthepony.mc.otm.client.render.RGBAColor;
import ru.dbotthepony.mc.otm.client.render.RenderHelper;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.List;

import static org.lwjgl.opengl.GL11.GL_ALWAYS;
import static org.lwjgl.opengl.GL11.GL_LESS;

public abstract class AbstractSlotPanel extends EditablePanel {
	public AbstractSlotPanel(@Nonnull MatteryScreen<?> screen, @Nullable EditablePanel parent, float x, float y, float width, float height) {
		super(screen, parent, x, y, width, height);
	}

	public AbstractSlotPanel(@Nonnull MatteryScreen<?> screen, @Nullable EditablePanel parent, float x, float y) {
		super(screen, parent, x, y, REGULAR_DIMENSIONS, REGULAR_DIMENSIONS);
	}

	public static final RGBAColor SLOT_HIGHLIGHT = new RGBAColor(255, 255, 255, 100);
	public static final RGBAColor SLOT_HIGHLIGHT_DRAG = new RGBAColor(200, 200, 200, 150);

	public static final int REGULAR_DIMENSIONS = 18;

	protected void renderSlotBackground(PoseStack stack, float mouse_x, float mouse_y, float flag) {
		RenderHelper.drawRegularSlot(stack, 0, 0);
	}

	protected void renderRegular(PoseStack stack, ItemStack itemstack) {
		renderRegular(stack, itemstack, null);
	}

	protected void renderRegular(PoseStack stack, ItemStack itemstack, @Nullable String count_override) {
		RenderSystem.setShader(GameRenderer::getPositionTexShader);

		float height = 0;

		if (!itemstack.isEmpty()) {
			RenderSystem.enableDepthTest();

			var system_stack = RenderSystem.getModelViewStack();

			system_stack.pushPose();
			system_stack.translate(parent_x + 1, parent_y + 1, 0);
			RenderSystem.applyModelViewMatrix();

			RenderSystem.depthFunc(GL_LESS);

			// Thanks Mojang
			// Very cool
			// (for int x, int y, which are then cast into doubles anyway)
			screen.getItemRenderer().blitOffset = accumulated_depth - 100; // force item to draw only 50 units "above" background
			screen.getItemRenderer().renderAndDecorateItem(Minecraft.getInstance().player, itemstack, 0, 0, (int) (parent_x + parent_y * 1000) /* some kind of indexing ? */);

			RenderSystem.depthFunc(GL_ALWAYS);
			screen.getItemRenderer().renderGuiItemDecorations(screen.getFont(), itemstack, 0, 0, count_override);

			screen.getItemRenderer().blitOffset = 0;

			// too big accumulations can lead to Z near clipping issues
			accumulate3DHeight(101);
			height = 101;

			system_stack.popPose();
			RenderSystem.applyModelViewMatrix();
		}

		if (is_hovered) {
			stack.pushPose();
			stack.translate(0, 0, height);
			RenderHelper.setDrawColor(SLOT_HIGHLIGHT);
			RenderHelper.drawRect(stack, 1, 1, getWidth() - 1, getHeight() - 1);
			stack.popPose();
		}
	}

	@Nonnull
	abstract protected ItemStack getItemStack();

	protected List<Component> getItemStackTooltip(ItemStack stack) {
		return screen.getTooltipFromItem(stack);
	}

	@Override
	protected void innerRender(PoseStack stack, float mouse_x, float mouse_y, float flag) {
		renderSlotBackground(stack, mouse_x, mouse_y, flag);

		renderRegular(stack, getItemStack());
	}

	@Override
	protected boolean innerRenderTooltips(PoseStack stack, float mouse_x, float mouse_y, float flag) {
		if (is_hovered) {
			var itemstack = getItemStack();

			if (!itemstack.isEmpty()) {
				Font font = net.minecraftforge.client.RenderProperties.get(itemstack).getFont(itemstack);
				// TODO: WHERE????????????
				// GuiUtils.preItemToolTip(itemstack);
				screen.renderComponentTooltip(stack, getItemStackTooltip(itemstack), (int) mouse_x, (int) mouse_y, (font == null ? screen.getFont() : font));
				// GuiUtils.postItemToolTip();

				return true;
			}
		}

		return false;
	}
}
