package ru.dbotthepony.mc.otm.menu.widget;

import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import ru.dbotthepony.mc.otm.capability.matter.IPatternStorage;
import ru.dbotthepony.mc.otm.menu.MatteryMenu;
import ru.dbotthepony.mc.otm.menu.data.LongDataContainer;

import javax.annotation.Nullable;
import java.util.List;
import java.util.function.Supplier;

public class PatternStorageWidget extends GaugeWidget {
	public Supplier<Long> value_supplier;
	public Supplier<Long> max_value_supplier;

	protected final LongDataContainer value_container = new LongDataContainer();
	protected final LongDataContainer max_value_container = new LongDataContainer();

	public PatternStorageWidget(MatteryMenu menu, int x, int y, IPatternStorage capability) {
		this(menu, x, y, () -> (long) capability.getStored(), () -> (long) capability.getCapacity());
	}

	public PatternStorageWidget(MatteryMenu menu, IPatternStorage capability) {
		this(menu, 0, 0, () -> (long) capability.getStored(), () -> (long) capability.getCapacity());
	}

	public PatternStorageWidget(MatteryMenu menu) {
		this(menu, 0, 0, null, null);
	}

	public PatternStorageWidget(MatteryMenu menu, int x, int y, Supplier<Long> value_supplier, Supplier<Long> max_value_supplier) {
		super(menu, x, y);
		this.value_supplier = value_supplier;
		this.max_value_supplier = max_value_supplier;

		addDataSlots(value_container);
		addDataSlots(max_value_container);
	}

	public PatternStorageWidget(MatteryMenu menu, Supplier<Long> value_supplier, Supplier<Long> max_value_supplier) {
		this(menu, 0, 0, value_supplier, max_value_supplier);
	}

	@Override
	public void updateServer() {
		if (value_supplier == null || max_value_supplier == null)
			return;

		value_container.setValue(value_supplier.get());
		max_value_container.setValue(max_value_supplier.get());
	}

	@Override
	public int getImageWidth() {
		return 9;
	}

	@Override
	public int getImageHeight() {
		return 48;
	}

	@Override
	public int getImageX() {
		return 0;
	}

	@Override
	public int getImageY() {
		return 148;
	}

	@Nullable
	@Override
	public List<Component> getTooltip() {
		return List.of(
			new TranslatableComponent("otm.gui.pattern.percentage_level", String.format("%.2f", getLevel() * 100d)),
			new TranslatableComponent("otm.gui.pattern.format", value_container.getValue(), max_value_container.getValue())
		);
	}

	@Override
	public float getLevel() {
		return max_value_container.getValue() == 0 ? 0 : (float) ((double) value_container.getValue() / (double) max_value_container.getValue());
	}
}
