package ru.dbotthepony.mc.otm.network.android;

import net.minecraft.client.Minecraft;
import net.minecraft.network.FriendlyByteBuf;
import net.minecraftforge.api.distmarker.Dist;
import net.minecraftforge.fml.DistExecutor;
import net.minecraftforge.network.NetworkEvent;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;
import ru.dbotthepony.mc.otm.capability.android.AndroidCapabilityPlayer;

import java.util.function.Supplier;

public record AndroidStatusPacket(Type type, boolean status) {
	public enum Type {
		IS_ANDROID,
		WILL_BECOME_ANDROID,
	}

	public void write(FriendlyByteBuf buffer) {
		buffer.writeByte(type.ordinal());
		buffer.writeBoolean(status);
	}

	public void play(Supplier<NetworkEvent.Context> context) {
		context.get().setPacketHandled(true);
		context.get().enqueueWork(() -> {
			DistExecutor.unsafeRunWhenOn(Dist.CLIENT, () -> this::playClient);
		});
	}

	public void playClient() {
		var ply = Minecraft.getInstance().player;

		if (ply != null) {
			ply.getCapability(MatteryCapability.ANDROID).ifPresent(cap -> {
				if (cap instanceof AndroidCapabilityPlayer pcap) {
					if (type == Type.IS_ANDROID)
						pcap.is_android = status;
					else if (type == Type.WILL_BECOME_ANDROID)
						pcap.will_become_android = status;
				}
			});
		}
	}

	public static AndroidStatusPacket read(FriendlyByteBuf buffer) {
		return new AndroidStatusPacket(Type.values()[buffer.readByte()], buffer.readBoolean());
	}
}
